/*
 * Copyright (c) 2022 Nordic Semiconductor ASA.
 *
 * SPDX-License-Identifier: LicenseRef-Nordic-5-Clause
 */

#include <psa/crypto.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tfm_secure_api.h"
#include "tfm_api.h"

#include "tfm_spm_log.h"

#include "nrfx_gpiote.h"
#include "hal/nrf_egu.h"
#include "hal/nrf_timer.h"
#include "hal/nrf_spim.h"

#include "psa/service.h"
#include "psa_manifest/tfm_secure_peripheral_partition.h"

#include "util.h"

#include <autoconf.h>
#include <zephyr/devicetree.h>

#include <nrfx_uarte.h>

#define BUTTON_PIN      DT_GPIO_PIN(DT_NODELABEL(button0), gpios)

#if defined(CONFIG_BOARD_NRF5340DK_NRF5340_CPUAPP_NS)
#define SCK_PIN         47 /* P1.15 */
#define MOSI_PIN        45 /* P1.13 */
#elif defined(CONFIG_BOARD_NRF9160DK_NRF9160_NS)
#define SCK_PIN         13 /* P0.13 */
#define MOSI_PIN        11 /* P0.11 */
#endif

#define TIMER_RELOAD_VALUE (1*1000 * 1000)
static uint32_t m_button_count;
static uint32_t m_timer_count;
static uint32_t m_trigger_count;
#define M_GPIOTE_CHANNEL 0

/** @brief Symbol specifying UARTE instance to be used. */
#define UARTE_INST_IDX 0

/** @brief Symbol specifying TX pin number of UARTE. */
#define UARTE_TX_PIN 20

/** @brief Symbol specifying RX pin number of UARTE. */
#define UARTE_RX_PIN 19

/** @brief Symbol specifying the first message to be sent via UARTE data transfer as part of @ref m_tx_buffer. */
#define MSG1 "Nordic"

/** @brief Symbol specifying the second message to be sent via UARTE data transfer as part of @ref m_tx_buffer. */
#define MSG2 " Semiconductor"

/** @brief Symbol specifying the third message to be sent via UARTE data transfer as part of @ref m_tx_buffer. */
#define MSG3 " nRF"

/**
 * @brief Structure containing receive buffers registers.
 *
 * @note To show the double-buffered functionality of data pointer register @p buff_2 is prepared to
 *       store @p MSG3 and @p buff_3 is prepared to store @p MSG2.
 */
typedef struct
{
    uint8_t buff_1[sizeof(MSG1)];   ///< RX buffer 1
    uint8_t buff_2[sizeof(MSG3)];   ///< RX buffer 2
    uint8_t buff_3[sizeof(MSG2)];   ///< RX buffer 3
} rx_buffers_t;

/**
 *  @brief UARTE transmit buffer, it is later filled with @ref MSG1, @ref MSG2 and @ref MSG3 to store
 *         following message: "Nordic Semiconductor nRF".
 */
static uint8_t m_tx_buffer[sizeof(MSG1) + sizeof(MSG2) + sizeof(MSG3)];

/** @brief Structure containing receive buffers. */
static rx_buffers_t m_rx_buffers;

static void timer_init(NRF_TIMER_Type *timer, uint32_t ticks)
{
	nrf_timer_mode_set(timer, NRF_TIMER_MODE_TIMER);
	nrf_timer_bit_width_set(timer, NRF_TIMER_BIT_WIDTH_32);
	nrf_timer_prescaler_set(timer, NRF_TIMER_FREQ_1MHz);
	nrf_timer_cc_set(timer, NRF_TIMER_CC_CHANNEL0, ticks);
	/* Clear the timer once event is generated. */
	nrf_timer_shorts_enable(timer, NRF_TIMER_SHORT_COMPARE0_CLEAR_MASK);
}

static void timer_stop(NRF_TIMER_Type *timer)
{
	nrf_timer_task_trigger(timer, NRF_TIMER_TASK_STOP);
	nrf_timer_int_disable(timer, NRF_TIMER_INT_COMPARE0_MASK);
	nrf_timer_event_clear(timer, NRF_TIMER_EVENT_COMPARE0);
}

static void timer_start(NRF_TIMER_Type *timer)
{
	timer_stop(timer);

	nrf_timer_task_trigger(timer, NRF_TIMER_TASK_CLEAR);
	nrf_timer_int_enable(timer, NRF_TIMER_INT_COMPARE0_MASK);

	nrf_timer_task_trigger(timer, NRF_TIMER_TASK_START);
}

static void timer_event_clear(NRF_TIMER_Type *timer)
{
	nrf_timer_event_clear(timer, NRF_TIMER_EVENT_COMPARE0);
}

psa_flih_result_t tfm_timer1_irq_flih(void)
{
	timer_event_clear(NRF_TIMER1);

	m_trigger_count++;

	if (m_trigger_count == 10) {
		m_trigger_count = 0;
		nrf_egu_task_trigger(NRF_EGU0_NS, NRF_EGU_TASK_TRIGGER0);
		return PSA_FLIH_SIGNAL;
	}

	return PSA_FLIH_NO_SIGNAL;
}

static void gpio_init(uint32_t pin)
{
	nrf_gpio_cfg_input(pin, NRF_GPIO_PIN_PULLUP);
	nrf_gpiote_event_configure(NRF_GPIOTE0, M_GPIOTE_CHANNEL, pin,
				   GPIOTE_CONFIG_POLARITY_HiToLo);
	nrf_gpiote_event_enable(NRF_GPIOTE0, M_GPIOTE_CHANNEL);
	nrf_gpiote_int_enable(NRF_GPIOTE0, NRFX_BIT(M_GPIOTE_CHANNEL));
}

psa_flih_result_t tfm_gpiote0_irq_flih(void)
{
	nrf_gpiote_event_clear(NRF_GPIOTE0, nrf_gpiote_in_event_get(M_GPIOTE_CHANNEL));

	nrf_egu_task_trigger(NRF_EGU0_NS, NRF_EGU_TASK_TRIGGER0);
	return PSA_FLIH_SIGNAL;
}

static void spim_init(uint32_t sck_pin, uint32_t mosi_pin)
{
	nrf_spim_pins_set(NRF_SPIM3, sck_pin, mosi_pin, NRF_SPIM_PIN_NOT_CONNECTED);
	nrf_spim_configure(NRF_SPIM3, NRF_SPIM_MODE_0, NRF_SPIM_BIT_ORDER_MSB_FIRST);
#if SPIM0_FEATURE_HARDWARE_CSN_PRESENT
	nrf_spim_csn_configure(NRF_SPIM3,
		NRF_SPIM_PIN_NOT_CONNECTED,
		NRF_SPIM_CSN_POL_LOW,
		0);
#endif
	nrf_spim_frequency_set(NRF_SPIM3, NRF_SPIM_FREQ_2M);
	nrf_spim_int_enable(NRF_SPIM3, NRF_SPIM_INT_ENDTX_MASK);
	nrf_spim_enable(NRF_SPIM3);
}

static void spim_send(const uint8_t *buf, size_t len)
{
	nrf_spim_tx_buffer_set(NRF_SPIM3, buf, len);
	nrf_spim_rx_buffer_set(NRF_SPIM3, NULL, 0);

	nrf_spim_task_trigger(NRF_SPIM3, NRF_SPIM_TASK_START);
}

static void spim_event_clear(void)
{
	nrf_spim_event_clear(NRF_SPIM3, NRF_SPIM_EVENT_ENDTX);
}

static size_t generate_msg(uint8_t *msg_buf, uint8_t msg_len,
			   uint32_t timer_count, uint32_t button_count)
{
	size_t hash_length;
	uint8_t secret[4 + 4];
	uint8_t hash[32];

	util_put_le32(timer_count, &secret[0]);
	util_put_le32(button_count, &secret[4]);

	psa_status_t status = psa_hash_compute(PSA_ALG_SHA_256,
				secret, sizeof(secret),
				hash, sizeof(hash), &hash_length);
	if (status != PSA_SUCCESS) {
		psa_panic();
	}

	return util_bin2hex(hash, hash_length, msg_buf, msg_len);
}

static void send_msg(void)
{
	static uint8_t msg_buf[2 * 32 + 1];
	size_t msg_len;

	msg_len = generate_msg(msg_buf, sizeof(msg_buf),
			       m_timer_count, m_button_count);
	psa_irq_enable(TFM_SPIM3_IRQ_SIGNAL);

	spim_send(msg_buf, msg_len);
	if (psa_wait(TFM_SPIM3_IRQ_SIGNAL, PSA_BLOCK) != TFM_SPIM3_IRQ_SIGNAL) {
		psa_panic();
	}

	spim_event_clear();
	psa_eoi(TFM_SPIM3_IRQ_SIGNAL);

	psa_irq_disable(TFM_SPIM3_IRQ_SIGNAL);
}

static void spp_signals_process(psa_signal_t signals)
{
	if (signals & TFM_TIMER1_IRQ_SIGNAL) {
		m_timer_count++;

		SPMLOG_INFMSGVAL("IRQ: TIMER1 count: ", m_timer_count);

		psa_reset_signal(TFM_TIMER1_IRQ_SIGNAL);
	}
	if (signals & TFM_GPIOTE0_IRQ_SIGNAL) {
		m_button_count++;

		SPMLOG_INFMSGVAL("IRQ: GPIOTE0 count: ", m_button_count);

		psa_reset_signal(TFM_GPIOTE0_IRQ_SIGNAL);
	}
}

static void spp_send(void)
{
	psa_status_t status;
	psa_msg_t msg;

	status = psa_get(TFM_SPP_SEND_SIGNAL, &msg);

	send_msg();

	psa_reply(msg.handle, status);
}

static void spp_signal_handle(psa_signal_t signals)
{
	psa_status_t status;
	psa_msg_t msg;

	status = psa_get(TFM_SPP_PROCESS_SIGNAL, &msg);

	spp_signals_process(signals);

	psa_reply(msg.handle, status);
}

static void spp_init(void)
{
	timer_init(NRF_TIMER1, TIMER_RELOAD_VALUE);
	timer_start(NRF_TIMER1);

	psa_irq_enable(TFM_TIMER1_IRQ_SIGNAL);

	gpio_init(BUTTON_PIN);

	psa_irq_enable(TFM_GPIOTE0_IRQ_SIGNAL);

	spim_init(SCK_PIN, MOSI_PIN);
}

/**
 * @brief Function for handling UARTE driver events.
 *
 * @param[in] p_event   Pointer to event structure. Event is allocated on the stack so it is available
 *                      only within the context of the event handler.
 * @param[in] p_context Context passed to the interrupt handler, set on initialization. In this example
 *                      p_context is used to pass the address of the UARTE instance that calls this handler.
 */
static void uarte_handler(nrfx_uarte_event_t const * p_event, void * p_context)
{
    nrfx_err_t status;
    (void)status;

    nrfx_uarte_t * p_inst = p_context;
    static uint8_t num_handler_exec = 0;

    switch (num_handler_exec++)
    {
        case 0:
            status = nrfx_uarte_rx(p_inst,
                                   m_rx_buffers.buff_3,
                                   NRFX_ARRAY_SIZE(m_rx_buffers.buff_3) - 1);
            NRFX_ASSERT(status == NRFX_SUCCESS);
            break;

        case 1:
            status = nrfx_uarte_rx(p_inst,
                                   m_rx_buffers.buff_2,
                                   NRFX_ARRAY_SIZE(m_rx_buffers.buff_2) - 1);
            NRFX_ASSERT(status == NRFX_SUCCESS);
            break;
    }

    if (p_event->type == NRFX_UARTE_EVT_TX_DONE)
    {
        printf("--> TX done\n");
        printf("--> Bytes transfered: %u\n", p_event->data.error.rxtx.bytes);
        nrfx_uarte_uninit(p_inst);
    }
}

/**
 * @brief Function for showing values of all receive buffers.
 *
 * @param[in] p_rx_buff Pointer to the structure with receive buffers.
 */
static void rx_buffers_print(rx_buffers_t * p_rx_buff)
{
    printf("................................");
    printf("RX buffer1 - addr: %p | content: %s\n", p_rx_buff->buff_1, p_rx_buff->buff_1);
    printf("RX buffer2 - addr: %p | content: %s\n", p_rx_buff->buff_2, p_rx_buff->buff_2);
    printf("RX buffer3 - addr: %p | content: %s\n", p_rx_buff->buff_3, p_rx_buff->buff_3);
    printf("................................");
}


/**
 * @brief Function for application main entry.
 *
 * @return Nothing.
 */
void uarte0_test()
{
    nrfx_err_t status;
    (void)status;


    printf("Starting nrfx_uarte RX double-buffered example.\n");

    nrfx_uarte_t uarte_inst = NRFX_UARTE_INSTANCE(UARTE_INST_IDX);
    nrfx_uarte_config_t uarte_config = NRFX_UARTE_DEFAULT_CONFIG(UARTE_TX_PIN, UARTE_RX_PIN);
    uarte_config.p_context = &uarte_inst;
    status = nrfx_uarte_init(&uarte_inst, &uarte_config, uarte_handler);
    NRFX_ASSERT(status == NRFX_SUCCESS);

#if defined(__ZEPHYR__)
    #define UARTE_INST         NRFX_CONCAT_2(NRF_UARTE, UARTE_INST_IDX)
    #define UARTE_INST_HANDLER NRFX_CONCAT_3(nrfx_uarte_, UARTE_INST_IDX, _irq_handler)
    IRQ_DIRECT_CONNECT(NRFX_IRQ_NUMBER_GET(UARTE_INST), IRQ_PRIO_LOWEST, UARTE_INST_HANDLER, 0);
#endif

    /* Declaration and filling m_tx_buffer to store desired msg ("Nordic Semiconductor nRF"). */
    memcpy(m_tx_buffer                              , MSG1, strlen(MSG1));
    memcpy(m_tx_buffer + strlen(MSG1)               , MSG2, strlen(MSG2));
    memcpy(m_tx_buffer + strlen(MSG1) + strlen(MSG2), MSG3, strlen(MSG3) + 1);
    printf("Content of TX buffer: %s\n", m_tx_buffer);

    rx_buffers_print(&m_rx_buffers);

    status = nrfx_uarte_rx(&uarte_inst,
                           m_rx_buffers.buff_1,
                           NRFX_ARRAY_SIZE(m_rx_buffers.buff_1) - 1);
    NRFX_ASSERT(status == NRFX_SUCCESS);

    status = nrfx_uarte_tx(&uarte_inst, m_tx_buffer, NRFX_ARRAY_SIZE(m_tx_buffer));
    NRFX_ASSERT(status == NRFX_SUCCESS);

    while (nrfx_uarte_tx_in_progress(&uarte_inst))
    {}

    rx_buffers_print(&m_rx_buffers);

    while (1)
    {
		psa_wait(PSA_WAIT_ANY, PSA_BLOCK);
    }
}



psa_status_t tfm_spp_main(void)
{
	psa_signal_t signals = 0;

	spp_init();

    uarte0_test();

	while (1) {
		signals = psa_wait(PSA_WAIT_ANY, PSA_BLOCK);
		if (signals & TFM_SPP_SEND_SIGNAL) {
			spp_send();
		} else if (signals & TFM_SPP_PROCESS_SIGNAL) {
			spp_signal_handle(signals);
		} else if (signals & (TFM_TIMER1_IRQ_SIGNAL |
				      TFM_GPIOTE0_IRQ_SIGNAL)) {
			/* Partition schedule was run while signals was set. */
			spp_signals_process(signals);
		} else {
			psa_panic();
		}
	}

	return PSA_ERROR_SERVICE_FAILURE;
}
